/*	analog.cpp
*
*	Game Timer battery monitor and voltage control
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*/

#include <avr/io.h>
#include "main.h"
#include "analog.h"

/*
REF V: 11 -> 1.1V internal; 01 -> AVcc
ADLAR: 1 -> left adjust
prescaler: 110 -> 8 Mhz / 64 = 125 kHz for ADC clock (8 ms)
  (13 ADC cycles for conversion = 104 ms)
  
note that we use 1.1V for battery measurement and Vcc for volume;
after switching the ref source, then next conversion may be innacurate, so should be discarded
*/

const uchar admuxBatt = 0xe6;	// 111. 0110 (1.1V ref, left adjust result [8 bits], channel 6)
const uchar admuxVol = 0x67;	// 011. 0111 (Vcc ref, left adjust result [8 bits], channel 7)

/*
*	battery voltage
*/

/* voltage at ADC input will be: rGnd/(rVcc+rGnd) * Vcc
   ADC result will be: Vin/1.1 * 255  [or Vin/11 * 255 if Vin is in decivolts]
*/
const uchar rVcc = 75;
const uchar rGnd = 24;
#define BattADCForDv(dv) ((long)rGnd * (dv) * 255 / ((rVcc + rGnd) * 11))

/* setup mode battery voltage meter */
const uchar dvMin = 30;		// decivolts when just 1 LED shown
const uchar dvMost = 45;	// decivolts when all 12 LEDs lit
const uchar battLow = BattADCForDv(dvMin);
const uchar dbatt = BattADCForDv(dvMost - dvMin);
const uchar cstepBatt = 11;		// 11 'steps' around circle for battery meter (1 always lit)

/* hysteresis band for low battery indicator */
const uchar dvHystHigh = 37;	// high value (decivolts) of hysteresis band; need to get over this to be 'good' (1.2V per cell)
const uchar dvHystLow = 35;		// low value (decivolts) of hysteresis band; battery is low when below this
	/* low happens about when we drop to 3 LEDs on setup mode gauge;
	   high is when we're nicely into 5 LEDs (36 would be 4, close to 5) */
const uchar battHystHigh = BattADCForDv(dvHystHigh);
const uchar battHystLow = BattADCForDv(dvHystLow);

/*
*	volume
*/
const uchar dvolTol = 0x03;		// tolerance for 'same'


ANALOG::ANALOG() : m_ccCur(ccNil), m_ccPrev(ccNil)
{
	ADCSRA = 0x86;	// 1000 0110 (enabled, no auto trigger, no interrupt, clk/64)
}

uchar ANALOG::FBattLevel(uchar *pbatt)
{
	uchar b;

	if (!FGetVal(ccBatt, &b))
		return fFalse;

	/* the number of steps to show is:
	
	   cur - low
	   --------- * steps
	   interval
	   
	   we add dbatt/2 to the numerator for rounding
	*/
	*pbatt = ((b - battLow) * cstepBatt + dbatt/2) / dbatt;
	return fTrue;
}

uchar ANALOG::FBattChange(uchar *pfBattLow)
{
	uchar b;
	
	if (!FGetVal(ccBatt, &b))
		return fFalse;

	/* apply hysteresis anaylsis so we don't wiggle */
	if (*pfBattLow)
	{
		if (b > battHystHigh)
		{
			*pfBattLow = fFalse;
			return fTrue;
		}
	}
	else
	{
		if (b < battHystLow)
		{
			*pfBattLow = fTrue;
			return fTrue;
		}
	}
	
	return fFalse;
}

uchar ANALOG::FVolChange(uchar *pvol)
{
	uchar b;
	
	if (!FGetVal(ccVol, &b))
		return fFalse;
	
	if ((*pvol > b ? *pvol - b : b - *pvol) > dvolTol)
	{
		*pvol = b;
		return fTrue;
	}
	return fFalse;
}

uchar ANALOG::FGetVal(uchar cc, uchar *pb)
{
	uchar fRet;

	*pb = 0;	// grrr... make compiler happy (uninit'd var warning in callers)
	
	if (ADCSRA & 0x40)	// conversion in progress
		return fFalse;	// nothing to report, and can't start another, yet
	
	fRet = fFalse;

	/* obviously need current result to be right, but also need prev to match since
	   battery and volume use different reference voltages and the first conversion
	   after switching reference voltage may be inaccurate */
	if (m_ccCur == cc && m_ccPrev == cc)
	{
		*pb = ADCH;
		fRet = fTrue;
	}
	
	/* start another going for next time */
	m_ccPrev = m_ccCur;
	m_ccCur = cc;
	ADMUX = cc == ccBatt ? admuxBatt : admuxVol;
	ADCSRA |= 0x40;		// start conversion
	
	return fRet;
}

